#!/usr/bin/env python3
#_*_ coding:utf-8 _*_

import ahocorasick
from .diPwm import diPWM
from .Enumerate import enumerate_words_LAM, enumerate_words_LAM_ratio

##################################################
def search_aho(diP, text, threshold):
	"""Find in the sequence <text>, all occurrences of the di-PWM <diP> whose score is greater than the minimum score threshold <threshold>
           1. generate the list of valid words for the di-PWM <diP> satisfying the score constraint and build an Aho-Corasick automaton for them
           2. search for occurrences of these valid words in <text> using the Aho-Corasick automaton and report the list of all their occurrences

         Other description: Search of a set of words through a text using Aho-Corasick algorithm for a given threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""

	# create empty automaton
	autom = ahocorasick.Automaton()

	# fill automaton with words and scores
	for word, score in enumerate_words_LAM(diP, threshold):
		autom.add_word(word, (word,score))
	autom.make_automaton()

	# search with the automaton through the text : yields starting position in the text, word, score
	for (position, (word, score)) in autom.iter(text):
		yield position-diP.length, word, score

##################################################
def search_aho_ratio(diP, text, ratio):
	""" Find occurrences of the di-PWM <diP> in the sequence <text> for a ratio threshold <ratio>
            1. computes a score threshold from the ratio
            2. call function search_aho to perform the search of the occurrences in the sequence

            Other description: 
              Search of a set of words through a text using Aho-Corasick algorithm for a given ratio.
              From the ratio is calculated the threshold.

	Args:
 		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		ratio (float): float or int. From 0 to 1

	Yields:
		tuple: starting_position, word, score
	"""
	threshold = diP.set_threshold_from_ratio(ratio)
	for position, word, score in search_aho(diP, text, threshold):
		yield position, word, score


##################################################
# fn added 10.02.23
# good default pvalue threshold ? here 10^-6

def search_aho_pvalue(diP, text, pvalue = 0.000001 ):
	""" Find occurrences of the di-PWM <diP> in the sequence <text> for the pvalue threshold <pvalue>
            1. computes a score threshold from the pvalue
            2. call function search_aho to perform the search of the occurrences in the sequence

        Remark: it is difficult to find a good pvalue threshold ; we advise to make several trials starting with very low (restrictive) threshold values. Otherwise, use the search_aho_ratio function instead.

	Args:
 		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		pvalue (float): float. pvalue threshold given to select the occurrences.
                                Between 0 and strictly less to 1

	Yields:
		tuple: starting_position, word, score
	"""
        
	# calculate threshold from pvalue
	threshold = diP.set_threshold_from_pvalue( pvalue )
        # perform the search
	for position, word, score in search_aho(diP, text, threshold):
		yield position, word, score




# SOL https://py-pkgs.org/04-package-structure.html
# read version from installed package
import sys
from importlib.metadata import version
__version__ = version("dipwmsearch")

# ##################################################
# print ( __version__ )
sys.stderr.write( f"version {__version__}\n"  )
# ##################################################

from .AhoCorasick import *
# from .BestWords import *
from .Block import *
from .diPwm import *
from .Enumerate import *
from .SemiNaive import *
# from .Super import *
# Q: do we need the following line?
# from .__main__ import *
#



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_

import statistics, sys
import math
import ahocorasick
from .diPwm import diPWM, create_diPwm
from .SemiNaive import search_semi_naive_LAM_window, search_semi_naive_LAM


def compute_positions_kept(diP, k):
	""" Compute the list of positions of the diPWM object where the standard deviation is more than `k`.

	Args:
		diP (diPWM): object diPWM

		k (float): threshold given to select the position of the diPWM object

	Return:
		list of int: a list of positions of the diPWM object
	"""

	list_positions_kept = []
	for i in range(diP.length):
		dico = diP.value[i]
		standard_dev = statistics.stdev(list(dico.values()))
		if standard_dev > k :
			list_positions_kept.append(i)
	# FR version # print(f'Liste des positions gardées : {list_positions_kept}', file=sys.stderr)
	print(f'List of kept positions : {list_positions_kept}', file=sys.stderr)
	return list_positions_kept


def make_blocks(list_positions_kept):
	""" Translate a list of position in a list of blocks.

	Args:
		list_positions_kept (list of int): list of positions

	Return:
		list of int: a list of blocks
	"""

	list_blocks = []
	for i in range(len(list_positions_kept)):
		if i == 0 or list_positions_kept[i] != list_positions_kept[i - 1] + 1:
			list_blocks.append([])
			list_blocks[-1].append(list_positions_kept[i])
		else:
			list_blocks[-1].append(list_positions_kept[i])
	# FR version # print(f'Liste des blocs : {list_blocks}', file=sys.stderr)
	print(f'List of blocks: {list_blocks}', file=sys.stderr)
	return list_blocks


def get_longest_block(list_blocks):
	""" Give the longest block of a list of blocks.

	Args:
		list_blocks (list of int): list of blocks

	Return:
		list of int: list of successive positions (a block)
	"""

	list_length = []
	for block in list_blocks:
		list_length.append(len(block))
	index_longest_block = (-max((x,-i) for i,x in enumerate(list_length))[1])
	# FR version # print(f'Plus long bloc : {list_blocks[index_longest_block]}, taille bloc : {len(block)}', file=sys.stderr)
	# ER comment: error when printing the length of the longest block. Currently block is the last block of the list, not the longest one. Corrected below 22.08.22.
	print(f'Longest block: {list_blocks[index_longest_block]},  bloc length: {list_length[index_longest_block]}', file=sys.stderr)
	return list_blocks[index_longest_block]


def enumerate_block(diP, block, threshold):
	""" This function enumerates words of length of `block + 1` which scores are in the total diPWM >= to `threshold`.
	To estimate score, this function uses LAM and LBM to extend the score on the right and on the left.

	Args:
		diP (diPWM): object diPWM

		block (list of int): list of successive positions (a block)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: word, score
	"""

	LBM = diP.make_look_back_matrix()
	stack = []

	# for root, child and parent : [0]:level, [1]:score, [2]:prefix
	# one root per symbol of the alphabet
	for nt in diP.alphabet:
		root = (block[0], LBM[block[0]][nt],nt)
		stack.append(root)

	while stack:
		parent = stack.pop()

		for d in diP.alphabet:
			child = (
				parent[0] + 1,
				parent[1] + diP.value[parent[0]][parent[2][-1] + d],
				parent[2] + d
			)
			if child[1] + diP.LAM[child[0]][d] >= threshold :
				if parent[0] == block[-1] :
					yield child[2], child[1] + diP.LAM[child[0]][d]		# generate word, score
				else :
					stack.append(child)

######################################################################
# search using the longest core block

def search_block(diP, text, threshold):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""

	list_positions_kept = compute_positions_kept(diP,1)
	list_blocks = make_blocks(list_positions_kept)
	block = get_longest_block(list_blocks)


	autom = ahocorasick.Automaton()
	for word, score in enumerate_block(diP, block, threshold):
		autom.add_word(word, (word,score))
	autom.make_automaton()

	for (position, (word, score)) in autom.iter(text):
		start_position = position - (block[-1] + 1)
		if position < len(text) - diP.length + block[-1]:
			score = search_semi_naive_LAM_window(diP, start_position ,text, threshold)
			if score:
				yield start_position,text[start_position:start_position+diP.length+1], score


######################################################################
# search_block_ratio
# takes a diP, a text, and a ratio:
# 1. convert the ratio in threshold and
# 2. call search_block with this threshold
                                
def search_block_ratio(diP, text, ratio):
	""" Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. From the ratio is calculated the threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""

	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for position, word, score in search_block(diP, text, threshold):
		yield position, word, score

############################################################
def get_block_optimized(diP):
	""" Compute a block of the diPWM object by a heuristic.

	Args:
		diP (diPWM): object diPWM

	Return:
		list of int: list of successive positions (a block)
	"""

    # Computes standard deviation for each position of the diPWM
	li_sd = [statistics.stdev(list(dico.values())) for dico in diP.value]

    # Computes mean of standard deviation for possible blocks
	li_interval_possible = [(list(range(i,j+1)),statistics.mean(li_sd[i:j+1])) for i in range(len(diP)) for j in range(i,len(diP)) if j - i + 1 >= min(len(diP)-3,10)]
	
    # Sort by decreasing value of mean
	li_interval_possible.sort(key=lambda x:x[1],reverse=True)

    # Return list of positions for the block with highest mean
	return li_interval_possible[0][0]

######################################################################
# similar to search_block but using the optimized core block

def search_block_optimized(diP, text, threshold):
	"""Find in the sequence <text>, all occurrences of the di-PWM <diP> whose score is greater than the minimum score threshold <threshold>. Done using a core block optimized strategy: the core block is a reduced di-PWM with selective positions. The search is done for the core block and core occurrences are extended to full length word occurrences. See the article for details.

           Other description: Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. The subwords are compute with a heuristic.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""
        # compute the core block of the di-PWM
        # this defines - a reduced di-PWM containing only the columns for core
        #              - and a core length that is shorter compared to the initial di-PWM
        
	block = get_block_optimized(diP)

        # enumerate the core valid words for the core length and build an Aho_Corasick automaton for them

	autom = ahocorasick.Automaton()
	for word, score in enumerate_block(diP, block, threshold):
		autom.add_word(word, (word,score))
	autom.make_automaton()
        
        # search for occurrences of core valid words
        # then check the window centered on the core valid word
        # to verify if there is a full-length occurrence

	for (position, (word, score)) in autom.iter(text):
		start_position = position - (block[-1] + 1)
		if position < len(text) - diP.length + block[-1]:
			score = search_semi_naive_LAM_window(diP, start_position ,text, threshold)
			if score:
				yield start_position,text[start_position:start_position+diP.length+1], score

############################################################
######################################################################
# search_block_optimized_ratio
# takes a diP, a text, and a ratio:
# 1. convert the ratio in threshold and
# 2. call search_block_optimized with this threshold
# remark: similar to search_block_ratio but with optimized core block

def search_block_optimized_ratio(diP, text, ratio):
	""" Find occurrences of the di-PWM <diP> in the sequence <text> for a ratio threshold <ratio>
            1. computes a score threshold from the ratio
            2. call function search_block_optimized to perform the search of the occurrences in the sequence

	Other description:
              Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. The subwords are compute with a heuristic. From the ratio is calculated the threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		threshold (float): threshold given to select the windows

	Yields:
		tuple: starting_position, word, score
	"""
	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for position, word, score in search_block_optimized(diP, text, threshold):
		yield position, word, score


######################################################################
# search_block_optimized_pvalue
# takes a diP, a text, and a pvalue threshold:
# 1. convert the pvalue threshold  in a score threshold and
# 2. call search_block_optimized with this threshold
# remark: similar to search_block_ratio but with a pvalue threshold instead
##########
# fn added 10.02.23
# good default pvalue threshold ? here 10^-6

def search_block_optimized_pvalue(diP, text, pvalue = 0.000001 ):
	""" Find occurrences of the di-PWM <diP> in the sequence <text>  for the pvalue threshold <pvalue>
            1. computes a score threshold from the pvalue
            2. call function search_block_optimized to perform the search of the occurrences in the sequence

        Remark: it is difficult to find a good pvalue threshold ; we advise to make several trials starting with very low (restrictive) threshold values. Otherwise, use the function search_block_optimized_ratio instead.

	Other description:
              Search of a set of subwords through a text using Aho-Corasick algorithm for a given threshold and check the windows by using LAM. The subwords are compute with a heuristic. From the ratio is calculated the threshold.

	Args:
		diP (diPWM): object diPWM

		text (string): text to search on the motif (first position = 0)

		pvalue (float): pvalue threshold given to select the occurrences.
                                Between 0 and strictly less to 1.

	Yields:
		tuple: starting_position, word, score
	"""
        
	# calculate threshold from pvalue
	threshold = diP.set_threshold_from_pvalue( pvalue )
        # perform the search
	for position, word, score in search_block_optimized(diP, text, threshold):
		yield position, word, score


#



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_


from .diPwm import diPWM, create_diPwm
import math


def search_semi_naive_LAM_window(diP, start_position, text, threshold):
    lengthSeqRef = len(text)
    score=0
    position_in_window = 0

    #Check for character IUPAC
    if text[start_position] not in diP.alphabet:
        return None
    # Check at each position of the window
    for j in range(start_position, start_position + diP.length):
        d = text[j]
        b = text[j + 1]

        #Check for character IUPAC
        if b not in diP.alphabet:
            return None

        score = score + diP.value[position_in_window][d + b]
        # print(j,score,score + diP.LAM[position_in_window + 1][b],diP.LAM[position_in_window + 1])
        # if partial score + estimate max score suffixe doesn't reach threshold : break
        if (score + diP.LAM[position_in_window + 1][b]) < threshold:

            return None
        # otherwise computation of partial score
        position_in_window += 1
    return score

def search_semi_naive_LAM(diP, text, threshold):
    """ Search of the diPWM through a text by sliding window for a given threshold.
    This function uses the LookAheadMatrix to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        threshold (float): threshold given to select the windows

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    lengthSeqRef = len(text)

    # Computation for every position of the text - size of diPWM
    for i in range(0, lengthSeqRef - 1 - diP.length):
        score=0
        complete = False

        #Check for character IUPAC
        if text[i] not in diP.alphabet:
            return None
        # Check at each position of the window
        for j in range(0, diP.length):
            d = text[i + j]
            b = text[i + j + 1]

            #Check for character IUPAC
            if b not in diP.alphabet:
                return None

            # if partial score + estimate max score suffixe doesn't reach threshold : break
            if (j<diP.length - 1) and ((score + diP.value[j][d + b] + diP.LAM[j + 1][b]) < threshold):
                score=(-math.inf)
                break
            # otherwise computation of partial score
            else:
                score = score + diP.value[j][d + b]
                if (j == diP.length - 1):
                    complete=True

        # if whole window score is >= threshold : yield starting position, sub-string, score
        if ((score >= threshold) and complete==True):
            yield i, text[i:i+diP.length+1], score


def search_semi_naive_LAM_ratio(diP, text, ratio):
    """Search of the diPWM through a text by sliding window for a given ratio.
    From the ratio is calculated the threshold.
    This function uses the LookAheadMatrix to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        ratio (float): float or int. From 0 to 1

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    threshold = diP.set_threshold_from_ratio(ratio)
    for position, mot, score in search_semi_naive_LAM(diP, text, threshold):
        yield position, mot, score


def search_semi_naive_LAT(diP, text, threshold):
    """ Search of the diPWM through a text by sliding window for a given threshold.
    This function uses the LookAheadTable to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        threshold (float): threshold given to select the windows

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    lengthSeqRef = len(text)
    LAT = diP.make_look_ahead_table()

    # Computation for every position of the text - size of diPWM
    for i in range(0,lengthSeqRef-1-diP.length):
        score=0
        complete = False

        #Check for character IUPAC
        if text[i] not in diP.alphabet:
            return None

        # Check at each position of the window
        for j in range(0,diP.length):
            d=text[i+j]
            b=text[i+j+1]

            #Check for character IUPAC
            if b not in diP.alphabet:
                return None

            # if partial score + estimate max score suffixe doesn't reach threshold : break
            if (j<diP.length-1) and ((score + diP.value[j][d + b] + LAT[j+1]) < threshold):
                score=(-math.inf)
                break
            # otherwise computation of partial score
            else:
                score = score + diP.value[j][d + b]
                if (j==diP.length-1):
                    complete=True

        # if whole window score is >= threshold : yield starting position, sub-string, score
        if ((score >= threshold) and complete==True):
            yield i, text[i:i+diP.length+1], score


def search_semi_naive_LAT_ratio(diP, text, ratio):
    """Search of the diPWM through a text by sliding window for a given ratio.
    From the ratio is calculated the threshold.
    This function uses the LookAheadTable to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        ratio (float): float or int. From 0 to 1

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    threshold = diP.set_threshold_from_ratio(ratio)
    for position, mot, score in search_semi_naive_LAT(diP, text, threshold):
        yield position, mot, score


def search_semi_naive_back_LBM(diP, text, threshold):
    """ Search of the diPWM through a text by sliding window for a given threshold.
    Each window is computed backward from right to left.
    This function uses the LookBackMatrix to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        threshold (float): threshold given to select the windows

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    lengthSeqRef = len(text)
    LBM= diP.make_look_back_matrix()

    # Computation for every position of the text starting at the length of the diPWM
    for i in range(diP.length, lengthSeqRef):
        score=0
        complete = False


        #Check for character IUPAC
        if text[i] not in diP.alphabet:
            return None

        # Check at each position of the window, starting from the end of the window
        # moving backward
        for j in range(diP.length-2,-2,-1):
            d=text[i-(diP.length-1-j)]
            b=text[i-(diP.length-1-j)+1]

            #Check for character IUPAC
            if b not in diP.alphabet:
                return None

            # if partial score + estimate max score prefix doesn't reach threshold : break
            if (j>0) and ((score + diP.value[j+1][d + b] + LBM[j][d]) < threshold):
                score=(-math.inf)
                break
            # otherwise computation of partial score
            else:
                score = score + diP.value[j+1][d + b]
                if (j==0):
                    complete=True

        # if whole window score is >= threshold : yield starting position, sub-string, score
        if ((score >= threshold) and complete==True):
            yield i-diP.length, text[i-(diP.length):i+1], score


def search_semi_naive_back_LBM_ratio(diP, text, ratio):
    """Search of the diPWM through a text by sliding window for a given ratio.
    From the ratio is calculated the threshold.
    Each window is computed backward from right to left.
    This function uses the LookBackMatrix to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        ratio (float): float or int. From 0 to 1

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    threshold = diP.set_threshold_from_ratio(ratio)
    for position, mot, score in search_semi_naive_back_LBM(diP, text, threshold):
        yield position, mot, score


def search_semi_naive_back_LBT(diP, text, threshold):
    """ Search of the diPWM through a text by sliding window for a given threshold.
    Each window is computed backward from right to left.
    This function uses the LookBackTable to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        threshold (float): threshold given to select the windows

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    lengthSeqRef = len(text)
    LBT = diP.make_look_back_table()

    # Computation for every position of the text starting at the length of the diPWM
    for i in range(diP.length, lengthSeqRef):
        score=0
        complete = False

        #Check for character IUPAC
        if text[i] not in diP.alphabet:
            return None

        # Check at each position of the window, starting from the end of the window
        # moving backward
        for j in range(diP.length-2,-2,-1):
            d=text[i-(diP.length-1-j)]
            b=text[i-(diP.length-1-j)+1]

            #Check for character IUPAC
            if b not in diP.alphabet:
                return None

            # if partial score + estimate max score prefix doesn't reach threshold : break
            if (j>0) and ((score + diP.value[j+1][d + b] + LBT[j]) < threshold):
                score=(-math.inf)
                break
            # otherwise computation of partial score
            else:
                score = score + diP.value[j+1][d + b]
                if (j==0):
                    complete=True

        # if whole window score is >= threshold : yield starting position, sub-string, score
        if ((score >= threshold) and complete==True):
            yield i-diP.length, text[i-(diP.length):i+1], score


def search_semi_naive_back_LBT_ratio(diP, text, ratio):
    """Search of the diPWM through a text by sliding window for a given ratio.
    From the ratio is calculated the threshold.
    Each window is computed backward from right to left.
    This function uses the LookBackTable to shorten computations.

    Args:
        diP (diPWM): object diPWM

        text (string): text to search on the motif (first position = 0)

        ratio (float): float or int. From 0 to 1

    Yields:
        tuple: starting position in the text, sub-string, score
    """
    threshold = diP.set_threshold_from_ratio(ratio)
    for position, mot, score in search_semi_naive_back_LBT(diP, text, threshold):
        yield position, mot, score



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_
from src.diPwm import diPWM, create_diPwm
from src.Enumerate import enumerate_words_LAM, enumerate_words_LAT, enumerate_words_LAM_ratio, enumerate_words_LAT_ratio
import pandas as pd

def test_enumerate_words_LAM():
	pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
	diP=create_diPwm(pathDiPwm)
	threshold = 20
	i=0
	for x in enumerate_words_LAM(diP,threshold):
		i+=1

	assert i == 41

def test_enumerate_words_LAT():
	pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
	diP=create_diPwm(pathDiPwm)
	threshold = 20
	i=0
	for x in enumerate_words_LAT(diP,threshold):
		i+=1

	assert i == 41

def test_enumerate_words_LAM_ratio():
	pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
	diP=create_diPwm(pathDiPwm)
	ratio = 0.95
	i=0
	for x in enumerate_words_LAM_ratio(diP,ratio):
		i+=1

	assert i == 158

def test_enumerate_words_LAT_ratio():
	pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
	diP=create_diPwm(pathDiPwm)
	ratio = 0.95
	i=0
	for x in enumerate_words_LAT_ratio(diP,ratio):
		i+=1

	assert i == 158


#!/usr/bin/env python3
#_*_ coding:utf-8 _*_
from src.diPwm import diPWM, create_diPwm
from src.Block import compute_positions_kept, make_blocks, get_longest_block, enumerate_block, search_block, get_block_optimized, search_block_optimized

def test_compute_positions_kept():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    list_positions_kept = compute_positions_kept(diPwm_matrix_new, 2)

    assert list_positions_kept == [1, 2, 3, 4, 6, 7]


def test_make_blocks():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    list_positions_kept = compute_positions_kept(diPwm_matrix_new, 2)
    list_blocks = make_blocks(list_positions_kept)

    assert list_blocks == [[1, 2, 3, 4], [6, 7]]


def test_get_longest_block():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    list_positions_kept = compute_positions_kept(diPwm_matrix_new, 2)
    list_blocks = make_blocks(list_positions_kept)
    block = get_longest_block(list_blocks)

    assert block == [1, 2, 3, 4]


def test_enumerate_block():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diPwm_matrix_new = create_diPwm(pathDiPwm)
    list_positions_kept = compute_positions_kept(diPwm_matrix_new, 2)
    list_blocks = make_blocks(list_positions_kept)
    block = get_longest_block(list_blocks)
    nb = 0
    for word, score in enumerate_block(diPwm_matrix_new, block, 10):
        nb += 1

    assert nb == 84


def test_search_block():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diP=create_diPwm(pathDiPwm)
    text = open("./data/sequence.txt").readlines()[0]
    threshold = 5
    nb = 0
    for starting_position, word, score in search_block(diP, text, threshold):
        nb += 1

    assert nb == 12


def test_get_block_optimized():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diP=create_diPwm(pathDiPwm)
    block = get_block_optimized(diP)

    assert block == [1, 2, 3, 4, 5, 6, 7, 8]


def test_search_block_optimized():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diP=create_diPwm(pathDiPwm)
    text = open("./data/sequence.txt").readlines()[0]
    threshold = 5
    nb = 0
    for starting_position, word, score in search_block_optimized(diP, text, threshold):
        nb += 1

    assert nb == 12



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_


"""Main Function of the module dipwmsearch
   search for occurrences of the di-PWM contained in file <pathDiPwm>
   in the sequence contained the FASTA file <pathFasta>
   using the threshold ratio <ratio> the default ratio is 0.95
   print occurrences found on standard output, in tabulated format
"""
 
import statistics, sys, os
import ahocorasick
from Bio import SeqIO
from Bio.Seq import Seq

from .diPwm import diPWM, create_diPwm
from .Block import search_block_optimized

######################################################################
# GLOBAL CONSTANTS 
RETURN_OK = 0
RETURN_BAD = 1

# #####################################################################
# Main function so that the module can be used as script

# dummy main function
def test_main() -> int:
    """Dummy main function (only for debug purposes)       
       Echo the input arguments to standard output
    """

    phrase = ""
   
    for i in range( len(sys.argv) ):
            phrase = phrase + " " + sys.argv[i]
    print(phrase)
    return RETURN_BAD

######################################################################
# default search function: simply called search
def search(pathDiPwm, pathFasta, ratio) -> int:

    """search for occurrences of the di-PWM contained in file <pathDiPwm>
       in the sequence contained the FASTA file <pathFasta>
       using the threshold ratio <ratio> the default ratio is 0.95
       """

    
    # Creation objet diPwm
    diP = create_diPwm(pathDiPwm)
    nameDiPwm = os.path.basename(pathDiPwm)
    sys.stderr.write( f'diPWM: {nameDiPwm}\n' )

    # Threshold défini en fonction du ratio
    threshold = diP.set_threshold_from_ratio(ratio)

    # Ouverture du fichier Fasta et conversion en objet seqIO
    file = open(pathFasta)
    seqRecord = SeqIO.read(file, "fasta")
    # DEBUG:
    # print( seqRecord.id, os.path.basename(pathFasta), os.path.basename(pathFasta)[:-3] )
    seq_name = os.path.basename(pathFasta)[:-3]
    # Convert sequence text in uppercase
    text = str(seqRecord.seq.upper())

    # Create seq Reverse Complement et convert in text
    seqRC = seqRecord.seq.reverse_complement()
    textRC = str(seqRC.upper())

    # Recherche dnas le brin +
    for position, word, score in search_block_optimized(diP, text, threshold):
        # print(f'{position} \t {score} \t +')
        print(f'{position}\t{word}\t{score}\t+')
    # Recherche dans le brin -
    for position, word, score in search_block_optimized(diP, textRC, threshold):
        # print(f'{position} \t {score} \t -')
        print(f'{position}\t{word}\t{score}\t-')

    return RETURN_OK

######################################################################
# top level code
# check command line parameters
# check file availability
# check threshold range
# execute a default search using function search

if __name__ == '__main__':
    
    sys.stderr.write(__doc__)

    # default threshold ratio value
    DEF_RATIO = 0.95
    # minimum threshold ratio value
    MIN_RATIO = 0.80
    # min number of arguments for this script (script name included)
    MIN_ARG = 3
    
    # process parameters
    argc = len(sys.argv)
    # check correct arguments
    if ( argc < MIN_ARG ):
        sys.stderr.write("Usage: {0}\n 1st argument : filename of a di-PWM\n 2nd argument: filename of a FASTA file\n 3rd argument: a threshold ratio (between 0.8 and 1; default: 0.95)\n".format( sys.argv[0] ) )
        exit( RETURN_BAD )
    else: 
        dipfile = sys.argv[1]
        seqfile = sys.argv[2]
        
    # second parameter : add visualisation of plots
    if ( argc > MIN_ARG ):
        ratio_thr = float(sys.argv[3])
        if (ratio_thr < MIN_RATIO):
            sys.stderr.write( f"Input ratio threshold {ratio_thr} is too low. This threshold must be in the range [0.8, 1]\n" )
            sys.exit( RETURN_BAD )
            
    else:
        ratio_thr = DEF_RATIO
        
    ##############################
    # test if dipfile is present and readable
    try:
        f = open( dipfile )
        f.close()
    except FileNotFoundError:
        sys.stderr.write( "Input file {0} cannot be found\n".format( dipfile ) )
        exit( RETURN_BAD )
    except PermissionError:
        sys.stderr.write( "You need read access to file {0}. Change your access rights and relaunch program.\n".format( dipfile ) )
        sys.exit( RETURN_BAD )

    ##############################
    # test if seqfile is present and readable
    try:
        f = open( seqfile )
        f.close()
    except FileNotFoundError:
        sys.stderr.write( "Input file {0} cannot be found\n".format( seqfile ) )
        exit( RETURN_BAD )
    except PermissionError:
        sys.stderr.write( "You need read access to file {0}. Change your access rights and relaunch program.\n".format( seqfile ) )
        sys.exit( RETURN_BAD )

    ## CONTROL MESSAGE with parameters
    sys.stderr.write( f'Search call: {dipfile} {seqfile} {ratio_thr}\n' )
    
    sys.exit( search( dipfile, seqfile, ratio_thr) )

# END of MAIN
######################################################################



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_
from src.diPwm import diPWM, create_diPwm
from src.AhoCorasick import search_aho
from math import fabs

DEBUG = False
cumulated_difference_threshold = 0.01

def test_pvalue():
    pathDiPwm = "./data/ATF4_HUMAN.H11DI.0.A.dpwm"
    diP=create_diPwm(pathDiPwm)

    # list of Pvalues thresholds
    L = [ 0.0001, 0.0005, 0.001 ]
    # empty list of score thresholds 
    S= []
    # precision parameters for Pvalue computation
    k = 0.025
    n = 2
    
    for i in L:
        score = diP.set_threshold_from_pvalue( i, k, n )
        print( i, " ", score )
        S.append( score )

    # Check whether values in S are close to those of solutions stored in Scores
    # CACCTGAAGCAATT 9.758137921158239 0.000150576234
    # TTAATAATGAAATG 4.636829825579451 0.000651195645
    # GTTCTGACGCATTG 2.0259309215181425 0.001265782863 
    Scores = [ 9.75813, 4.63682, 2.02593 ]
    diff = 0
    for i,j in zip(Scores,S):
        print( i-j, i, j )
        diff += fabs(i-j)

    print( "The cumulated difference between computed score thresholds and true thresholds is ", diff )
    # test with assert instruction
    assert  diff < cumulated_difference_threshold , "The cumulated difference between computed score thresholds and true thresholds is larger than {0}\n".format( cumulated_difference_threshold )



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_
from src.diPwm import diPWM, create_diPwm
from src.AhoCorasick import search_aho

def test_search_aho():
    pathDiPwm = "./data/ATF3_HUMAN.H11DI.0.A.dpwm"
    diP=create_diPwm(pathDiPwm)
    text = open("./data/sequence.txt").readlines()[0]
    threshold = 5
    nb = 0
    for starting_position, word, score in search_aho(diP, text, threshold):
        nb += 1

    assert nb == 12



#!/usr/bin/env python3
#_*_ coding:utf-8 _*_

from .diPwm import diPWM


def enumerate_words_LAM(diP, threshold):
	"""This function enumerates words of length of diPWM + 1 which scores are >= to threshold.
	To estimate score, this function uses LAM.

	Args:
		diP (diPWM): Object diPWM

		threshold (float): score of the words >= threshold will be enumerated

	Yields:
		tuple: word, score
	"""
	stack = []

	# for root, child and parent : [0]:level, [1]:score, [2]:prefix
	# one root per symbol of the alphabet
	for nt in diP.alphabet:
		root = (0,0,nt)
		stack.append(root)

	while stack:
		parent = stack.pop()

		for d in diP.alphabet:
			child = (
				parent[0] + 1,
				parent[1] + diP.value[parent[0]][parent[2][-1] + d],
				parent[2] + d
			)

			# case of the last step to create word of full length
			if parent[0] == diP.length-1:

				if child[1] >= threshold:
					yield child[2], child[1]	# generate word, score

			# other cases : inside the tree
			# create child if estimate score can reach threshold using LAM
			elif child[1] + diP.LAM[child[0]][d] >= threshold:

				#Child added as last element of the stack
				stack.append(child)

def enumerate_words_LAM_ratio(diP, ratio):
	"""This function enumerates words of length of diPWM + 1 which scores are >= to a threshold.
	Threshold is calculated from score range of the diPWM and ratio.
	To estimate score, this function uses LAM.

	Args:
		diP (diPWM): object diPWM

		ratio (float): float or int. From 0 to 1

	Yields:
		tuple: word, score
	"""
	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for mot, score in enumerate_words_LAM(diP, threshold):
		yield mot, score



def enumerate_words_LAT(diP, threshold):
	"""This function enumerates words of length of diPWM + 1 which scores are >= to threshold.
	To estimate score, this function uses LAT.

	Args:
		diP (diPWM): Object diPWM

		threshold (float): score of the words >= threshold will be enumerated

	Yields:
		tuple: word, score
	"""

	LAT = diP.make_look_ahead_table()
	stack = []

	# for root, child and parent : [0]:level, [1]:score, [2]:prefix
	# one root per symbol of the alphabet
	for nt in diP.alphabet:
		root = (0,0,nt)
		stack.append(root)

	while stack:
		parent = stack.pop()

		for d in diP.alphabet:
			child = (
				parent[0] + 1,
				parent[1] + diP.value[parent[0]][parent[2][-1] + d],
				parent[2] + d
			)

			# case of the last step to create word of full length
			if parent[0] == diP.length-1:

				if child[1] >= threshold:
					yield child[2], child[1]	# generate word, score

			# other cases : inside the tree
			# create child if estimate score can reach threshold using LAM
			elif child[1] + LAT[0] >= threshold:

				#Child added as last element of the stack
				stack.append(child)


def enumerate_words_LAT_ratio(diP, ratio):
	"""This function enumerates words of length of diPWM + 1 which scores are >= to a threshold.
	Threshold is calculated from score range of the diPWM and ratio.
	To estimate score, this function uses LAT.

	Args:
		diP (diPWM): object diPWM

		ratio (float): float or int. From 0 to 1

	Yields:
		tuple: word, score
	"""

	# calculate threshold from ratio
	threshold = diP.set_threshold_from_ratio(ratio)
	for mot, score in enumerate_words_LAT(diP, threshold):
		yield mot, score
